/*
 * Module that forms part of a controllable faulty network router
 *
 * Completed as part of a final year project in the
 *  Department of Computer Science,
 *  University of Wales, Aberstwyth
 *
 * (C) 2003 Thomas Salmon, this code is GNU GPL
 */

#include <linux/ip_faulty_router.h>

#include <linux/module.h>

#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_tables.h>

#include <linux/random.h>

/* struct sent from user space */
struct params {
	int rule_type; /* packet order of packet drop */
	int operation; /* append, insert, update, delete, flush */
	int proto; /* transport layer protocol */
	int rule_num; /* only used for insert, update and delete operations */
	int min_size; /* min size of packet to match */
	int max_size; /* max size of packet to match */
	int max_burst_size; /* packet reordering only!, max no of packets in burst*/
	int max_delay_size; /* packet reordering only!, max packets to delay by */
	int burst_skew; /* percentage skew */
	int delay_skew; /* percentage skew */
	int probability; /* chance in 1000 */
};

struct rules *root_rule = NULL;

/* copy values contained in structs */
void copy_rule(struct rules *from, struct rules *to){
		to->rule_type = from->rule_type;
		to->proto = from->proto;
		to->min_size = from->min_size;
		to->max_size = from->max_size;
		to->probability = from->probability;
		to->max_burst_size = from->max_burst_size;
		to->max_delay_size = from->max_delay_size;
		to->burst_skew = from->burst_skew;
		to->delay_skew = from->delay_skew;
}

/* desends down the list of rules, printing each one out */
void list_rules(struct rules *root){
	if (root == NULL){
		return;
	}

	/* display this rule */
	if (root->rule_type == DROP){
		printk("DROP    ");
	}
	else{
		printk("REORDER    ");
	}

	if (root->proto == PROTO_ALL){
		printk("-p ALL    ");
	}
	else if (root->proto == PROTO_TCP){
		printk("-p TCP    ");
	}
	else if (root->proto == PROTO_UDP){
		printk("-p UDP    ");
	}
	else if (root->proto == PROTO_ICMP){
		printk("-p ICMP    ");
	}
	else if (root->proto == PROTO_IGMP){
		printk("-p IGMP    ");
	}

	printk("min: %d  max: %d    ", root->min_size, root->max_size);

	if (root->rule_type == REORDER){
		printk("Burst: %d (%d)  ", root->max_burst_size, root->burst_skew);
		printk("Delay: %d (%d)    ", root->max_delay_size, root->delay_skew);
	}

	printk("Prob: %d\n", root->probability);

	/* go to display next rule */
	if (root->next != NULL){
		list_rules(root->next);
	}
	return;
}

/* adds a rule to the list of rules */
/* if position specified, rule inserted before that position */
void add_rule(struct rules *root, struct rules *rule, int position){
	if (root == NULL){
		return;
	}
	else if (root->next == NULL){
		root->next = rule;
	}
	else if (position == 0){ /* insert rule here */
		struct rules *r = kmalloc(sizeof(struct rules), GFP_ATOMIC);
		copy_rule(root, r);
		r->next = root->next;
		r->reorder = root->reorder;
		copy_rule(rule, root);
		root->next = r;
		root->reorder = NULL;

		kfree(rule); /* no longer needed */
	}
	else{
		position--;
		add_rule(root->next, rule, position); /* recurrsive call */
	}
}

struct rules *get_rule(struct rules *root, int position){
	if (root == NULL || position < 0) return NULL;
	if (position == 0) return root;
	if (root->next != NULL) {
		position--;
		return get_rule(root->next, position);
	}
	return NULL; /* position beyond length of list */
}

void delete_rule(struct rules *root, int position){
	/* cannot work in following conditions */
	if (root == NULL) return;
	if (root->next == NULL) return;
	if (position < 1) return;

	/* we delete the following rule, not this one! */
	if (position == 1){
		struct rules *r = root->next->next;
		kfree(root->next);
		root->next = r;
	}
	else{
		position--;
		delete_rule(root->next, position);
	}
}

void flush_packets(struct stored_packet *root){
	if (root == NULL) return; /* shouldnt have been called */

	/* sanity check */
	if (root->delayed_skb != NULL && root->delayed_info != NULL){
		/* reinject the stored packets, dont just waste them*/
		nf_reinject(root->delayed_skb, root->delayed_info, NF_ACCEPT);
	}

	if (root->next != NULL){
		flush_packets(root->next);
		kfree(root->next);
		root->next = NULL;
	}
	return;
}

void flush_reorder(struct reorder *root){
	if (root == NULL) return; /* shouldnt have been called */
	if (root->next != NULL){
		flush_reorder(root->next);
		kfree(root->next);
		root->next = NULL;
	}
	if (root->reorder != NULL){
		flush_packets(root->reorder);
		kfree(root->reorder);
		root->reorder = NULL;
	}
	return;
}

void flush_rules(struct rules *root){
	if (root == NULL) return; /* shouldnt have been called */
	if (root->next != NULL){
		flush_rules(root->next);
		kfree(root->next);
		root->next = NULL;
	}
	if (root->reorder != NULL){
		flush_reorder(root->reorder);
		kfree(root->reorder);
		root->reorder = NULL;
	}
	return;
}

/* obeys instructions received from user space */
int rules_operation(struct params *rule){
	struct rules *add = NULL;

	if (rule->operation == LIST){
		/* list all stored rules */
		if (root_rule != NULL){
			list_rules(root_rule);
		}
		return 0;
	}
	else if (rule->operation == FLUSH){
		if (root_rule != NULL){
			flush_rules(root_rule);
			kfree(root_rule);
			root_rule = NULL;
		}
		return 0;
	}
	else if (rule->operation == DELETE){
		if (root_rule == NULL) return -1;
		if (rule->rule_num == 0){
			/* special case to delete the root node */

			if (root_rule->next == NULL){
				kfree(root_rule);
				root_rule = NULL;
			}
			else {
				struct rules *r = root_rule->next;
				kfree(root_rule);
				root_rule = r;
			}
		}
		else{
			delete_rule(root_rule, rule->rule_num);
		}
		return 0;
	}

	/* simple sanity checks */
	if (rule->min_size < 0 || rule->min_size > rule->max_size) return -1;
	if (rule->probability <= 0 || rule->probability > 1000) return -1;
	if (rule->rule_type == REORDER){
		if (rule->max_burst_size <= 0 || rule->max_delay_size <= 0) return -1;
	}

	/* if rule is append, insert or update, form the rule here */
	if (rule->operation == APPEND 
 	 || rule->operation == INSERT 
	 || rule->operation == UPDATE){
		add = kmalloc(sizeof(struct rules), GFP_ATOMIC);

		add->rule_type = rule->rule_type;
		add->proto = rule->proto;
		add->min_size = rule->min_size;
		add->max_size = rule->max_size;
		if (rule->rule_type == REORDER){
			add->max_burst_size = rule->max_burst_size;
			add->max_delay_size = rule->max_delay_size;
			add->burst_skew = rule->burst_skew;
			add->delay_skew = rule->delay_skew;
		}
		else{
			add->max_burst_size = -1;
			add->max_delay_size = -1;
			add->burst_skew = 0;
			add->delay_skew = 0;
		}
		add->probability = rule->probability;
		add->reorder = NULL;
		add->next = NULL;
	}

	if (rule->operation == APPEND){
		if (root_rule == NULL) root_rule = add;
		else add_rule(root_rule, add, -1);
	}
	else if (rule->operation == INSERT){
		if (root_rule == NULL) root_rule = add;
		else add_rule(root_rule, add, rule->rule_num);
	}
	else if (rule->operation == UPDATE){
		struct rules *u = get_rule(root_rule, rule->rule_num);

		if (u != NULL){
			/* copy contents over */
			copy_rule(add , u);
		}
		kfree(add); /* free memory */

		if (u == NULL){
			return -1;
		}
	}
	else return -1; /* no such operation */
	return 0;
}

/* return a random int between 1 and max, with the skew applied as percentage */
int random_number(int max, int skew){
	unsigned int rand;
	int res;
	int tmp;
	get_random_bytes(&rand, sizeof(unsigned int));
	tmp = (((int)rand%max)/2) + (max/2); /* number produced is -max < x < +max*/
	res = tmp + ((skew*max)/100);
	if (res <= 0) return 1;
	if (res > max) return max;
	return res;
}

/* checks the probability of event occuring: 0-false, 1-true */
int chkprob(int prob){
	return (random_number(1000,0) < prob);
}



/****
 * function for dropping
 *****************/

/*
 * Traverses the rules, to see if this packet should be dropped
 * returns true (1) to drop packet
 */
int drop_traverser(struct rules *root, int proto, int len){
	if (root == NULL) return 0;
	if (root->rule_type != DROP) return drop_traverser(root->next, proto, len);
	if (root->proto > PROTO_ALL && root->proto != proto) /* wrong protocol */
		return drop_traverser(root->next, proto, len);
	if (root->min_size > len || root->max_size < len) /* wrong size */
		return drop_traverser(root->next, proto, len);

	/* the packet is the same protocol, and fits the length constraints of
	   this rule */
	if (chkprob(root->probability)) return 1;
	else return drop_traverser(root->next, proto, len);
}






/**********************************************************************
 *
 * Function that packets out of the normal order
 * stores them in link lists
 * creates reordering chains, when a rule is being applied (perhaps being
 *                                              applied several times at once 
 * 
 ***********************************************************************/

/* creates a new struct reorder, that forms part of the linked list */
struct reorder *add_reorder_chain(struct rules *root,
                                  struct reorder *r, 
								  struct stored_packet *p){
	/* sanity check */								
	if (root == NULL || p == NULL){
printk("ERROR: root or packet is null\n");		
		return NULL;
	}

	if (r != NULL){
		if (r->next == NULL) r->next = add_reorder_chain(root, NULL, p);
	}
	else if (r != NULL && r->next != NULL) {
		add_reorder_chain(root, r->next, p);
	}
	else{
		struct reorder *ro = kmalloc(sizeof(struct reorder), GFP_ATOMIC);
		ro->chosen_burst_size = random_number(root->max_burst_size, 
											 root->burst_skew);
		ro->delayed_packets = 1;
		ro->chosen_delay_size = random_number(root->max_delay_size, 
											 root->delay_skew);
		ro->current_delay = 0;
		ro->reorder = p;
		if (p->delayed_skb == NULL || p->delayed_info == NULL){
printk("Packet invalid\n");
			return NULL; /* failure */
		}
		p->next = NULL; /* initialise the following packet ptr to NULL */
		ro->next = NULL;
		return ro;
	}
	return r;
}

/* adds a packet to link list, returning its address */
struct stored_packet *add_packet(struct stored_packet *sp, 
                                 struct sk_buff *skb, 
								 struct nf_info *nf){
	if (sp == NULL){
		struct stored_packet *p = kmalloc(sizeof(struct stored_packet), 
												 GFP_ATOMIC);
		/*p->delayed_skb = pskb_copy(skb, GFP_ATOMIC);
		p->delayed_skb = skb_copy(skb, GFP_ATOMIC);
		*/
		p->delayed_skb = skb;
		p->delayed_info = nf;
		p->next = NULL;
		return p;
	}
	else{
		sp->next = add_packet(sp->next, skb, nf);
		return sp;
	}
}

/*
 * Traverses a chain of reordering operations currently being performed, that
 *are all descended from the same rule
 *Returns 1 if packet is not being delayed
 */
int traverse_reorder_chains(struct reorder *ro, 
                            struct sk_buff *skb, 
							struct nf_info *nf){
	if (ro == NULL) return 1;
	if (ro->delayed_packets < ro->chosen_burst_size){
		ro->delayed_packets++;
		/* add packet */
		ro->reorder = add_packet(ro->reorder, skb, nf);
		return 0;
	}
	if (ro->next == NULL) return 1;
	return traverse_reorder_chains(ro->next, skb, nf);
}

/*
 * Traverses the rules regarding reordering
 * Returns true (1) if the packet should be reinjected (not being reordered)
 */
int reorder_traverser(struct rules *root, int proto, int len, 
                       struct sk_buff *skb, struct nf_info *nf){
	if (root == NULL) return 1;
	if (root->rule_type != REORDER)
		return reorder_traverser(root->next, proto, len, skb, nf);
	if (root->proto > PROTO_ALL && root->proto != proto) /* wrong protocol */
		return reorder_traverser(root->next, proto, len, skb, nf);
	if (root->min_size > len || root->max_size < len) /* wrong size */
		return reorder_traverser(root->next, proto, len, skb, nf);

	/* this rule can be applied to the packet */
	if (chkprob(root->probability)){ /* forming new reorder chain */
		struct stored_packet *p = kmalloc(sizeof(struct stored_packet), 
										 GFP_ATOMIC);
		/*p->delayed_skb = pskb_copy(skb, GFP_ATOMIC);*/
		/*p->delayed_skb = skb_copy(skb, GFP_ATOMIC);*/
		p->delayed_skb = skb;
		p->delayed_info = nf;
		p->next = NULL;
		root->reorder = add_reorder_chain(root, root->reorder, p);
	}
	else if (traverse_reorder_chains(root->reorder,skb,nf)) 
		return reorder_traverser(root->next, proto, len, skb, nf);
		/* try and fill any correct reordering going on */
		/* else check next rule */

		/* packet was not delayed on any reorder chains */
	else return 0; /* packet has been taken out of order */
	return 0;
}




/**********************************************************************
 *
 * Function that take reordered packets and reinject them, as appropriate 
 * 
 ***********************************************************************/

/* called to reinject all packets that have been taken out of order */
void reinject_packets(struct stored_packet *p){
	if (p == NULL) return;
	if (p->delayed_skb != NULL && p->delayed_info != NULL){
		nf_reinject(p->delayed_skb, p->delayed_info, NF_ACCEPT);
	}
	if (p->next == NULL) return; /* stop here */
	reinject_packets(p->next);
	kfree(p->next);
	p->next = NULL;
}

/* assess and updates all the reorder which is under way
 * called by rule_update every time any packet arrives
 * Returns: 1 - if reorder job complete (delete it) */
int reorder_update(struct reorder *r){
	int completed = 0;
	if (r == NULL) return 0;
	if (r->delayed_packets >= r->chosen_burst_size){
		if (r->current_delay >= r->chosen_delay_size){
			if (r->reorder != NULL){
				reinject_packets(r->reorder);
				kfree(r->reorder);
				r->reorder = NULL;
			}
			completed = 1; /* this reorder process is finished */
		}
		else{
			r->current_delay++; /* increment how long they've been delayed */
		}
	}
	if (r->next != NULL){
		if (reorder_update(r->next)){
			/* the next reordering terminated, erase it from linked list */
			struct reorder *tmp = r->next->next;
			kfree(r->next);
			r->next = tmp;
		}
	}
	return completed;
}

/***********************************************************
 * SHOULD BE CALLED FOR EVERY PACKET THAT ARRIVES
 ***********************************************************/
/* scans through the rules, checking any reordering that is going on */
void rule_update(struct rules *root){
	if (root == NULL) return;
	if (root->reorder != NULL){
		if (reorder_update(root->reorder)){
			/* the root reorder process is completed, delete it! */

			if (root->reorder != NULL){
				if (root->reorder->next != NULL){
					struct reorder *tmp;
					tmp = root->reorder->next;
					kfree(root->reorder);
					root->reorder = tmp;
				}
				else{
					kfree(root->reorder);
					root->reorder = NULL;
				}
			}
		}
	}
	if (root->next != NULL) rule_update(root->next);
}





/*
 * First function called when packets arrive
 * Returns:
 *        NF_DROP - to discard
 *        NF_QUEUE - to pass on the packets
 */
static unsigned int dropper(unsigned int hook, 
                            struct sk_buff **skb, 
							const struct net_device *in, 
							const struct net_device *out){
	if (root_rule == NULL) return NF_QUEUE; /* there are no rules */
	if (drop_traverser(root_rule, skb[0]->h.ipiph->protocol, 
	   skb[0]->len)) return NF_DROP;
	return NF_QUEUE; /* packet should be passed on */
}


/*
 * Function called for packets that have 'NF_QUEUE' returned for them
 * To allow packets to pass, nf_reinject must be called for them
 */
void reorder(struct sk_buff *skb, struct nf_info *nf, void *data){
	if (reorder_traverser(root_rule, skb->h.ipiph->protocol, skb->len, skb, nf))
		nf_reinject(skb, nf, NF_ACCEPT);
	rule_update(root_rule);
}


/*
 * User space communication function
 */
int user_control(struct sock *sk, int cmd, void *user, unsigned int len){
	if (!capable(CAP_NET_ADMIN)) return -EPERM;
	rules_operation( ((struct params *)user) );
	return 0;
}
int user_control_signed(struct sock *sk, int cmd, void *user, int *len){
	return user_control(sk, cmd, user, (unsigned int)*len);
}

/* structs for registring components */
static struct nf_hook_ops drop_ops = {
	{ NULL, NULL },
	(nf_hookfn *)dropper,
	PF_INET,
	NF_IP_FORWARD,
	NF_IP_PRI_FILTER
};

static struct nf_sockopt_ops user_comms = {
	{ NULL, NULL },
	PF_INET,
    IPT_BASE_CTL+2, /* 66 */
    IPT_BASE_CTL+3, /* 67 */
    user_control, /* function to be called */
    IPT_BASE_CTL+2, /* 66 */
    IPT_BASE_CTL+3, /* 67 */
    user_control_signed, /* function to be called */
	0,
	NULL
};

static int __init init(void){
	printk("Faulty Network Router - Thomas Salmon (tws9@aber.ac.uk)\n");

	if (nf_register_hook(&drop_ops) < 0){
		printk(KERN_ERR "Failed to register hook for dropping\n");
		return -1;
	}
	if (nf_register_queue_handler(PF_INET, (nf_queue_outfn_t)reorder, NULL) < 0){
		printk(KERN_ERR "Failed to register queue for reordering\n");
		return -1;
	}
	if (nf_register_sockopt(&user_comms) < 0){
		printk(KERN_ERR "Failed to register communication socket with user\n");
		return -1;
	}
	return 0;
}

static void __exit fini(void){
	/* dispose of memory */
	if (root_rule != NULL){
		flush_rules(root_rule);
		kfree(root_rule);
	}

	/* unregister any hooks */
	nf_unregister_queue_handler(PF_INET);
	nf_unregister_hook(&drop_ops);
}

module_init(init);
module_exit(fini);
