/*
 * User space program to communicate with kernel module: ip_faulty_router.o
 *enabling the faulty parameters inside the module
 *
 * Author: Thomas Salmon tws9@aber.ac.uk
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "kernel_comms.h"
#include <linux/netfilter_ipv4/ip_tables.h>
#include "faultctrl.h"

void usage(){
	printf("FaultCTRL: Tool to control Faulty Network Router\n\nUSAGE:\n");
	printf("faultctrl \t -A drop|reorder [options] probability \n");
	printf("faultctrl \t -U num drop|reorder [options] probability \n");
	printf("faultctrl \t -I [num] drop|reorder [options] probability \n");
	printf("faultctrl \t -D num\n");
	printf("faultctrl \t -F\n");
	printf("faultctrl \t -L\n\n");
	printf("Rules:\n");
	printf("\t -A \t Append a rule to the current rules\n");
	printf("\t -U \t Update a rule specified by the rule number\n");
	printf("\t -D \t Delete a rule that matches the options\n");
	printf("\t -I \t Insert a rule at given position in chain of rules\n");
	printf("\t -F \t Flush all rules, clear all faulty routing\n");
	printf("\t -L \t List all rules in operation\n\n");
	printf("Options:\n");
	printf("\t -p protocol \t Transport Layer Protocol \n");
	printf("\t  \t Values: igmp, icmp, tcp, udp \n");
	printf("\t -min size \t Minimum size (bytes) of packet to be delayed\n");
	printf("\t -max size \t Maximum size (bytes) of packet to be delayed\n");
	printf("\t -maxburst size \t Maximum no. packets to be reordered\n");
	printf("\t -maxdelay size \t Maximum no. packets to delay by\n");
	printf("\t -burstskew skew\t Percentage skew to be applied to the burst\n");
	printf("\t -delayskew skew\t Percentage skew to be applied to the delay\n");
}


/* fault will contain all parameters for this rule */
struct params *fault;

/* performs sanity checks on *fault, return 1 if they pass */
int check_params(){
	if (fault->operation == FLUSH) return 1; /* special case */
	if (fault->operation == LIST) return 1; /* special case */

	if ((fault->operation == UPDATE || fault->operation == DELETE) 
		&& fault->rule_num < 0) return -1;
	if (fault->operation == DELETE) return 1; /* special case */

	if (!(fault->rule_type == DROP || fault->rule_type == REORDER)) return -1;
	if (fault->operation < APPEND || fault->operation > FLUSH) return -1;
	if ((fault->operation == UPDATE || fault->operation == DELETE) 
		&& fault->rule_num < 0) return -1;
	if (fault->min_size < 0 || fault->max_size < 0) return -1;
	if (fault->min_size >= fault->max_size) return -1;
	if (fault->max_burst_size < 0) return -1;
	if (fault->max_delay_size < 0) return -1;
	if (fault->probability < 1 || fault->probability > 1000) return -1;
	return 1; /* sanity checked */
}

/* outputs values stored in params struct */
void output_debugging(){
	printf("Rule Type: \t%d\n", fault->rule_type);
	printf("Operation: \t%d\n", fault->operation);
	printf("Proto: \t%d\n", fault->proto);
	printf("Rule num: \t%d\n", fault->rule_num);
	printf("Min size: \t%d\n", fault->min_size);
	printf("Max size: \t%d\n", fault->max_size);
	printf("Max burst: \t%d\n", fault->max_burst_size);
	printf("Max delay: \t%d\n", fault->max_delay_size);
	printf("Burst skew: \t%d\n", fault->burst_skew);
	printf("Delay skew: \t%d\n", fault->delay_skew);
	printf("Probability: \t%d\n", fault->probability);
}


/* passes params to kernel */
int pass_to_kernel(){
	int fd;
	if (check_params() == -1) return -1;

	/* connect to kernel module, ip_faulty_router */
	fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	return setsockopt(fd, 0, IPT_BASE_CTL+2, fault, sizeof(struct params));

	output_debugging();

	printf("\nMODULE IS NOT COMMUNICATING WITH KERNEL\n");
	return -1;
}

/* load the default values into the params */
void load_defaults(){
	fault->rule_type = DROP;
	fault->operation = APPEND;
	fault->proto = PROTO_ALL;
	fault->rule_num = -1;
	fault->min_size = 0;
	fault->max_size = 1500;
	fault->max_burst_size = 0;
	fault->max_delay_size = 0;
	fault->burst_skew = 0;
	fault->delay_skew = 0;
	fault->probability = 0;
}

int main(int argc, char **argv){
	int current_arg = 2;
	if (argc < 2){
		usage();
		return -1;
	}

	fault = malloc(sizeof(struct params));
	load_defaults();

	if (strcmp(argv[1], "-A") == 0 && argc >= 4){ /* append */
		fault->operation = APPEND;
	}
	else if (strcmp(argv[1], "-U") == 0 && argc >= 5){ /* update */
		fault->operation = UPDATE;
		fault->rule_num = atoi(argv[2]);
		current_arg++;
	}
	else if (strcmp(argv[1], "-I") == 0 && argc >= 4){ /* insert */
		fault->operation = INSERT;
		if (strcmp(argv[2], "drop") != 0 && strcmp(argv[2], "reorder") != 0){
			/* next arg is the rule_num */
			fault->rule_num = atoi(argv[2]);
			current_arg++;
		}
	}
	else if (strcmp(argv[1], "-D") == 0 && argc >= 3){ /* delete */
		fault->operation = DELETE;
		fault->rule_num = atoi(argv[2]);
		if (pass_to_kernel() != 0){
			printf("Error when deleting rule\n");
			return -1;
		}
		return 0;
	}
	else if (strcmp(argv[1], "-F") == 0){
		fault->operation = FLUSH;
		if (pass_to_kernel() != 0){
			printf("Error when flushing rules\n");
			return -1;
		}
		return 0;
	}
	else if (strcmp(argv[1], "-L") == 0){
		fault->operation = LIST;
		if (pass_to_kernel() != 0){
			printf("Error when flushing rules\n");
			return -1;
		}
		return 0;
	}
	else{
		usage();
		return -1;
	}

	/* read in the drop or reorder instruction */
	if (strcmp(argv[current_arg],"drop") == 0){
		fault->rule_type = DROP;
	}
	else if (strcmp(argv[current_arg],"reorder") == 0){
		fault->rule_type = REORDER;
	}
	else{
		usage();
		return -1;
	}
	current_arg++;

	/* read in the options */
	while (current_arg < argc-2){ /* -2 as all options are pairs */
		if (strcmp(argv[current_arg], "-p") == 0){
			if (strcmp(argv[current_arg+1], "icmp") == 0){
				fault->proto = PROTO_ICMP;
			}
			else if (strcmp(argv[current_arg+1], "igmp") == 0){
				fault->proto = PROTO_IGMP;
			}
			else if (strcmp(argv[current_arg+1], "tcp") == 0){
				fault->proto = PROTO_TCP;
			}
			else if (strcmp(argv[current_arg+1], "udp") == 0){
				fault->proto = PROTO_UDP;
			}
			else {
				usage();
				return -1;
			}
		}
		else if (strcmp(argv[current_arg], "-min") == 0){
			fault->min_size = atoi(argv[current_arg+1]);
		}
		else if (strcmp(argv[current_arg], "-max") == 0){
			fault->max_size = atoi(argv[current_arg+1]);
		}
		else if (strcmp(argv[current_arg], "-maxburst") == 0){
			fault->max_burst_size = atoi(argv[current_arg+1]);
		}
		else if (strcmp(argv[current_arg], "-maxdelay") == 0){
			fault->max_delay_size = atoi(argv[current_arg+1]);
		}
		else if (strcmp(argv[current_arg], "-burstskew") == 0){
			fault->burst_skew = atoi(argv[current_arg+1]);
		}
		else if (strcmp(argv[current_arg], "-delayskew") == 0){
			fault->delay_skew = atoi(argv[current_arg+1]);
		}
		else{
			usage();
			return -1;
		}
		current_arg = current_arg + 2; /* increment by 2, as options in pairs */
	}

	if (current_arg != argc-1){
		/* current_arg should be pointing to the last arg */
		usage();
		return -1;
	}

	/* read the probability, last arg */
	fault->probability = atoi(argv[argc-1]);

	if (fault->rule_type == REORDER && (fault->max_burst_size == 0 || fault->max_delay_size == 0)){
		usage();
		printf("\nNB: Please specify a max burst rate and delay size for reordring\n");
		return -1;
	}

	if (check_params() < 0){
		usage();
	}
	else if (pass_to_kernel() != 0) 
		printf("Error communicating with kernel module\n");

	free(fault);
	return 0;
}

